import 'package:flutter/material.dart';
import 'package:forme/forme.dart';
import 'package:forme/src/decorator/forme_decorator.dart';

typedef EmptyChecker<T> = bool Function(T value);
typedef ValueCounter<T> = int Function(T value);

class FormeInputDecorationDecorator<T> extends FormeDecorator<T> {
  final InputDecoration? decoration;
  final EmptyChecker<T>? emptyChecker;
  final InputCounterWidgetBuilder? buildCounter;
  final int? maxLength;
  final ValueCounter<T>? valueCounter;

  FormeInputDecorationDecorator({
    this.decoration,
    this.emptyChecker,
    this.buildCounter,
    this.maxLength,
    this.valueCounter,
  });

  @override
  Widget build(BuildContext context, Widget child) {
    InputDecoration decoration = _getEffectiveDecoration(context);
    bool isValidateQuietly = Forme.of(context)!.quietlyValidate;
    FormeFieldState<T> state = FormeField.of(context)!;
    if (!isValidateQuietly) {
      decoration =
          _getEffectiveDecoration(context).copyWith(errorText: state.errorText);
    }
    return InputDecorator(
      decoration: decoration,
      isFocused: state.hasFocus,
      child: child,
    );
  }

  InputDecoration _getEffectiveDecoration(BuildContext context) {
    final MaterialLocalizations localizations =
        MaterialLocalizations.of(context);
    final ThemeData themeData = Theme.of(context);
    FormeFieldState<T> state = FormeField.of(context)!;
    bool enabled = state.enabled;
    final InputDecoration effectiveDecoration =
        (decoration ?? const InputDecoration())
            .applyDefaults(themeData.inputDecorationTheme)
            .copyWith(
              enabled: enabled,
              hintMaxLines: decoration?.hintMaxLines,
            );

    if (effectiveDecoration.counter != null ||
        effectiveDecoration.counterText != null) {
      return effectiveDecoration;
    }

    if (buildCounter != null) {
      Widget? counter;
      final int currentLength = _getValueLength(state.value);
      final bool isFocused = state.hasFocus;
      final Widget? builtCounter = buildCounter!(
        context,
        currentLength: currentLength,
        maxLength: maxLength,
        isFocused: isFocused,
      );
      if (builtCounter != null) {
        counter = Semantics(
          container: true,
          liveRegion: isFocused,
          child: builtCounter,
        );
      }
      return effectiveDecoration.copyWith(counter: counter);
    }

    if (maxLength == null) {
      return effectiveDecoration;
    }

    int currentLength = _getValueLength(state.value);

    String counterText = '$currentLength';
    String semanticCounterText = '';

    // Handle a real maxLength (positive number)
    if (maxLength! > 0) {
      // Show the maxLength in the counter
      counterText += '/$maxLength';
      final int remaining = (maxLength! - currentLength)
          .clamp(0, maxLength!); // ignore_clamp_double_lint
      semanticCounterText =
          localizations.remainingTextFieldCharacterCount(remaining);
    }

    bool hasIntrinsicError = currentLength > maxLength!;

    if (hasIntrinsicError) {
      return effectiveDecoration.copyWith(
        errorText: effectiveDecoration.errorText ?? '',
        counterStyle: effectiveDecoration.errorStyle ??
            (themeData.useMaterial3
                ? _m3CounterErrorStyle(context)
                : _m2CounterErrorStyle(context)),
        counterText: counterText,
        semanticCounterText: semanticCounterText,
      );
    }

    return effectiveDecoration.copyWith(
      counterText: counterText,
      semanticCounterText: semanticCounterText,
    );
  }

  int _getValueLength(T value) {
    return (valueCounter ?? _valueCounter).call(value);
  }

  int _valueCounter(T value) {
    if (value is String) {
      return value.length;
    }
    if (value is Iterable) {
      return value.length;
    }
    if (value is Map) {
      return value.length;
    }
    throw 'can not get the length of value $value';
  }

  TextStyle _m2CounterErrorStyle(BuildContext context) => Theme.of(context)
      .textTheme
      .bodySmall!
      .copyWith(color: Theme.of(context).colorScheme.error);

  TextStyle _m3CounterErrorStyle(BuildContext context) => Theme.of(context)
      .textTheme
      .bodySmall!
      .copyWith(color: Theme.of(context).colorScheme.error);
}
